MODULE MakeMake;

  (*
   * MAKEMAKE.  Create a MAKEFILE for a MODULA-2 program.
   *
   * Written by Steve Tynor, 30 September 1986.
   *            UUCP  : tynor@gitpyr
   *            USNAIL: 2550 Akers Mill Rd. T-2, Atlanta GA. 30339
   *
   * Permission is granted to distribute, copy and change this program as long
   * as this notice remains...
   *)

  (*
   * Modified for Logitech MODULA-2/86 by
   *           Robert A. Dukelow, 24 January 1986
   *           3467 Mt. Ariane Dr.
   *           San Diego, CA 92111
   *           UUCP: dukelow@nosc
   *
   * Program is set up for my working environment which has the compiler and
   * linker on drive D:. Most people will probably have to change
   * CompilerPathname and LinkerPathname.
   *
   * Primary changes:
   *   1. Changed philosophy when .DEF/.MOD files not found in local directory.
   *      Assume that we are *not* trying to maintain these files with this
   *      makefile. Include corresponding .SYM/.LNK files only in dependancies
   *      for other files and attempt to find the correct paths using the
   *      M2LNK/M2SYM environment variables.
   *   2. Uses CmdEnv written by Gregory Elder to both get the command line
   *      arguments and the environment variables since Logitech doesn't have
   *      a CommandLine module. (note that I had to Modify CmdEnv to allow for
   *      my expanded environment)
   *   3. The output file is now named after the main program module with
   *      a .MAK extension instead of MAKEFILE to allow maintaining more than
   *      one program in a directory.
   *   4. Creates an EXE file in addition to a LOD.
   *   5. Removed VAR's on parameters where I thought they were not necessary
   *      (mostly to reduce my own confusions).
   *
   *  Along the way I learned some things about Modula-2 portability:
   *   1. I discovered a couple of uninitialized variables in the original
   *      code. Maybe the TDI Modula-2 initializes memory?
   *   2. Seems that TDI allows passing a string constant as the actual
   *      parameter when parameter is declared as VAR? Don't know what it
   *      would do if you tried to change the value.
   *   3. TDI has a Strings module *very* similar to Logitech's (but not
   *      quite the same - sigh).
   *        a. Strings.Compare => Strings.CompareStr   (no big deal)
   *        b. Logitech uses +1/0/-1 vs Strings.?/Equal/Less
   *                                                   (easy to figure out)
   *        c. Strings.Pos has same name but different number of arguments
   *                                     (I guessed right the first time :-)
   *        d. Strings.Assign has the arguments reversed (ugh - debug time)
   *   4. ASCII.CR/LF/HT => ASCII.cr/lf/ht  (no real problem)
   *   5. TDI Modula apparently doesn't require EXPORTs in the DEF files.
   *  Overall - I sure would like to see more standardization.
   *
   * (see below for original authors comments)
   *)

  (*
   * MAKEMAKE.  Create a MAKEFILE for a MODULA-2 program.
   * Usage:
   *      MAKEMAKE "main-module-modulename"
   *
   *    Even though module dependencies are explicit in MODULA-2, it can still
   * be quite a chore to figure them out and create a MAKEFILE.  This program 
   * reads the text file associated with the module name given on the command 
   * line and recursively checks dependencies for all imported files.  It then
   * creates a MAKEFILE suitable for use with the Unix MAKE utility.  If a 
   * file is not found and is not a recognized library module, then the
   * MAKEFILE will include a comment:
   *     ## File NOT Found ##
   * on the right hand side of the dependency.
   *
   *
   * This program was written on the Atari ST using TDI Modula-2. 
   *
   * BUGS:
   *     the CommandLine module dies when a null command line is seen.  
   * Unfortunately, I no longer have the source to the module, so I can't
   * fix it.  This bug also seems to consider a command line consisting of
   * a single character as a reason for dieing.
   *
   * Suggestions for porting to another compiler:
   *
   *    1) module CommandLine is non-standard.  It exports 2 procedures:
   *          PROCEDURE GetArgument (    num      : CARDINAL; 
   *                                 VAR filename : ARRAY OF CHAR);
   *          PROCEDURE NumberOfArguments() : CARDINAL;
   *       they do just what their names imply. (ie. like C's argv, argc)
   *
   *    2) the IMPLEMENTATION MODULE MakeLibraries will have to be edited to
   *       reflect the standard library that your compiler knows about.  List
   *       any modules that you don't want included in your MAKEFILE.  Library
   *       modules are listed in the MAKEFILE, but are commented out.
   *
   *    3) the procedure ModulenameToFilename in MakeParse will have to be
   *       modified so that it creates the correct filename for your machine.
   *
   *    4) if your MAKE program has default MODULA-2 rules, set the constant
   *       MAKEHasModulaRules (in this module) to TRUE.  This will prevent
   *       MAKEMAKE from producing remake lines. 
   *    
   *    5) the procedure WriteFilename takes a parameter: extension.  Trace
   *       all the calls to this function and make sure the proper extentions
   *       are used.  in TDI Modula, '.DEF' = definition module text
   *                                 '.MOD' = implementation module text
   *                                 '.SYM' = symbol file (compiled DEF)
   *                                 '.LNK' = symbol file (compiled MOD)
   *)

FROM SYSTEM IMPORT BYTE;
IMPORT Debug;
IMPORT Break;
IMPORT InOut;
IMPORT CmdEnv;
IMPORT MakeForest;
IMPORT MakeParse;
IMPORT ASCII;
IMPORT FileSystem;
IMPORT Strings;

CONST
  MAKEHasModulaRules = FALSE;
  CompilerPathname   = 'm2 d:comp ';
  CompilerOptions    = '/b';
  LinkerPathname     = 'm2 d:link ';

VAR
  outputStream : FileSystem.File;
  column: CARDINAL; (* Output file column *)
  indent: CARDINAL; (* amount to indent when extending lines *)

  (*----------------------------------------------------------------------*)
  PROCEDURE CloseOutput;
  VAR
    reply : INTEGER;
  BEGIN
    FileSystem.Close(outputStream);
  END CloseOutput;

  (*----------------------------------------------------------------------*)
  PROCEDURE OpenOutput(): BOOLEAN;
  VAR
    filename: ARRAY [0..11] OF CHAR;
    ColPeriod: CARDINAL;
  BEGIN
    (* create a "makefile" with a name of the form mainmodule.MAK *)
    ColPeriod := Strings.Pos('.', MakeForest.FileForestEnd^.filename);
    Strings.Copy(MakeForest.FileForestEnd^.filename, 0, ColPeriod, filename);
    Strings.Concat(filename, '.MAK', filename);

    (* delete any existing file since FileSystem won't truncate old file *)
    FileSystem.Delete(filename, outputStream);

    (* create new output file *)
    FileSystem.Lookup(outputStream, filename, TRUE);

    column := 0;
    RETURN outputStream.res = FileSystem.done;
  END OpenOutput;

  (*----------------------------------------------------------------------*)
  PROCEDURE Write (ch : CHAR);
  BEGIN
    FileSystem.WriteByte (outputStream, BYTE(ch));
    INC(column);
  END Write;


  (*----------------------------------------------------------------------*)
  PROCEDURE WriteString (str : ARRAY OF CHAR);
  VAR
    c : CARDINAL;
  BEGIN
    c := 0;
    WHILE (c <= HIGH(str)) AND (str[c] <> 0C) DO
      Write(str[c]);
      INC(c);
    END; (* WHILE *)
  END WriteString;


  (*----------------------------------------------------------------------*)
  PROCEDURE WriteLn;
  BEGIN
    Write ( ASCII.cr );
    Write ( ASCII.lf );
    column := 0;
  END WriteLn;


  (*----------------------------------------------------------------------*)
  PROCEDURE SetIndent(i: CARDINAL);
  BEGIN
    indent := i;
  END SetIndent;


  (*----------------------------------------------------------------------*)
  PROCEDURE DoIndent;
  VAR
    i: CARDINAL;
  BEGIN
    FOR i := 1 TO indent DO Write(' ') END;
  END DoIndent;


  (*----------------------------------------------------------------------*)
  PROCEDURE WriteMakefile;

    (*--------------------------------------------------------------------*)
     PROCEDURE WriteFileDependency (file : MakeForest.FileDefinitionList;
                                    lhsExtention : ARRAY OF CHAR);
     VAR
       iptr : MakeForest.ImportList;
       TmpStr: ARRAY[0..100] OF CHAR;
     BEGIN
       IF file <> NIL THEN
         IF file^.filename[0] = '#' THEN
           RETURN;
         END; (* IF *)
         IF (NOT file^.local) OR (file^.path[0] <> 0C) OR file^.library THEN
           HALT; (* double check for debugging *)
         END;
         WriteLn;
         WriteFilename (file^.filename, lhsExtention);
         WriteString (': ');
         SetIndent(Strings.Length(file^.filename) + 1);
         WriteString (file^.filename);
           (* if this is a .LNK file, then don't forget the .SYM file: *)
           (* but don't do it for the main module! *)
         IF (file <> MakeForest.FileForestEnd) AND
            (Strings.CompareStr (lhsExtention, '.LNK') = 0 (* Strings.Equal *)) THEN
           Write (' ');
           WriteFilename (file^.filename, '.SYM');
         END; (* IF *)
         iptr := file^.imports;
         WHILE iptr <> NIL DO
           IF NOT iptr^.file^.library THEN
             MakeFullFilename(iptr^.file^.path, iptr^.file^.filename, '.SYM',
                              TmpStr);
             IF (column + Strings.Length(TmpStr) + 3) > 79 THEN
               WriteString (' \');
               WriteLn;
               DoIndent;
             END; (* IF *)
             Write (' ');
             WriteString(TmpStr);
           END; (* IF *)
           iptr := iptr^.next;
         END; (* WHILE *)

         iptr := file^.imports;
         column := 1000;  (* <-- force a WriteLn *)
         WHILE iptr <> NIL DO
           IF iptr^.file^.library THEN
             MakeFullFilename(iptr^.file^.path, iptr^.file^.filename, '.SYM',
                              TmpStr);
             IF (column + Strings.Length(TmpStr) + 1) > 79 THEN
               WriteLn;
               WriteString ('#libs:  ');
             END; (* IF *)
             Write (' ');
             WriteString(TmpStr);
           END; (* IF *)
           iptr := iptr^.next;
         END; (* WHILE *)

         WriteLn;
         IF NOT MAKEHasModulaRules THEN
           Write (ASCII.ht);
           WriteString (CompilerPathname);
           WriteString (file^.filename);
           WriteString (CompilerOptions);
           WriteLn;
         END; (* IF *)
       END; (* IF *)
     END WriteFileDependency;

    (*--------------------------------------------------------------------*)
    PROCEDURE WriteAllLnkFiles;
    VAR
      ptr : MakeForest.FileDefinitionList;
      TmpStr: ARRAY[0..100] OF CHAR;
    BEGIN
      ptr := MakeForest.FileForest;
      REPEAT
        IF Strings.Pos('.MOD', ptr^.filename) <= HIGH(ptr^.filename) THEN
          IF NOT ptr^.library THEN
            MakeFullFilename(ptr^.path, ptr^.filename, '.LNK', TmpStr);
            IF (column + Strings.Length(TmpStr) + 3) > 79 THEN
              WriteString (' \');
              WriteLn;
              DoIndent;
            END; (* IF *)
            Write (' ');
            WriteString(TmpStr);
          END; (* IF *)
        END; (* IF *)
        ptr := ptr^.next;
      UNTIL ptr = NIL;
      WriteLn;
    END WriteAllLnkFiles;

    (*--------------------------------------------------------------------*)
    PROCEDURE MakeFullFilename (path, filename, extention : ARRAY OF CHAR;
                                VAR fullname: ARRAY OF CHAR);
    VAR
      c : CARDINAL;
    BEGIN
      c := 0;
      REPEAT
        fullname[c] := filename[c];
        INC(c);
      UNTIL (c > HIGH (filename)) OR (filename[c] = '.') OR (filename[c] = 0C);
      fullname[c] := 0C;
      Strings.Concat(path, fullname, fullname);
      Strings.Concat(fullname, extention, fullname);
    END MakeFullFilename;

    (*--------------------------------------------------------------------*)
    PROCEDURE WriteFilename (filename, extention : ARRAY OF CHAR);
    VAR
      c : CARDINAL;
    BEGIN
      c := 0;
      REPEAT
        Write (filename[c]);
        INC(c);
      UNTIL (c > HIGH (filename)) OR (filename[c] = '.') OR (filename[c] = 0C);
      WriteString (extention);
    END WriteFilename;

    (*--------------------------------------------------------------------*)
    PROCEDURE WriteLODDependency (file : MakeForest.FileDefinitionList);
    BEGIN

      IF (NOT file^.local) OR (file^.path[0] <> 0C) OR file^.library THEN
        HALT; (* double check for debugging *)
      END;

      (* this is special stuff that didn't seem worth generalizing *)
      WriteLn;
      WriteString('# The following converts to EXE file');
      WriteLn;
      WriteFilename (file^.filename, '.EXE: ');
      WriteFilename (file^.filename, '.LOD');
      WriteLn;
      Write(ASCII.ht);
      WriteString('m2 d:lod2exe ');
      WriteFilename (file^.filename, '.LOD');
      WriteLn;

      WriteLn;
      WriteFilename (file^.filename, '.LOD');
      WriteString (':');
      SetIndent(Strings.Length(file^.filename) + 1);
      WriteAllLnkFiles;
      IF NOT MAKEHasModulaRules THEN
        Write (ASCII.ht);  WriteString (LinkerPathname);
        WriteFilename (file^.filename, '.LNK');
        WriteLn;
      END; (* IF *)
    END WriteLODDependency;

  VAR
    ptr    : MakeForest.FileDefinitionList;
(*  c      : CARDINAL; *)
  BEGIN
    IF OpenOutput() THEN
        (* print out the .LOD dependency first.... *)
      WriteLODDependency (MakeForest.FileForestEnd);
      WriteFileDependency (MakeForest.FileForestEnd, '.LNK');
      ptr := MakeForest.FileForest;
      WHILE ptr <> MakeForest.FileForestEnd DO
        IF (NOT ptr^.library) AND ptr^.local THEN
          IF Strings.Pos ('.MOD', ptr^.filename) <= HIGH(ptr^.filename) THEN
            WriteFileDependency (ptr, '.LNK');
          ELSE
            WriteFileDependency (ptr, '.SYM');
          END; (* IF *)
        END; (* IF *)
        ptr     := ptr^.next;
      END; (* WHILE *)
    ELSE
      InOut.WriteString ('Error opening MAKEFILE');
      InOut.WriteLn;
    END; (* IF *)
    CloseOutput;
  END WriteMakefile;

VAR
  modulename,
  filename : ARRAY [0 .. 132] OF CHAR;

BEGIN
  IF CmdEnv.ArgC < 1 THEN  
    InOut.WriteString ("Usage:   MAKEMAKE main-module-modulename");
    InOut.WriteLn;
  ELSE
    Strings.Assign(CmdEnv.ArgV[1]^, modulename);
    MakeParse.ParseModule (modulename, MakeParse.main, filename);
    WriteMakefile;
  END; (* IF *)
END MakeMake.